Notebook[{
Cell[
"
This file was created automatically by vma2nb from an .vma file.
It should be suitable to be loaded by Mathematica's GUI front
end but you probably don't want to edit it; edit the corresponding
.vma source file instead!
"
, "Text", Evaluatable->False
]

,Cell["\n\n  teleporter.vma   (C) 1999, 2000  Timo Felbinger\n\n  A demonstration of the quantum teleportation protocol for qubits.\n  (C. H. Bennett et. al. (1995))\n ","Text", Evaluatable->False ]
,Cell["
 SetOptions[ $Output, PageWidth -> 72 ];", "Input"]
,Cell["
 << qmatrix.m", "Input"]
,Cell["\n Define the actors: Alice and Bob: ","Text", Evaluatable->False ]
,Cell["\n Alice has two qubits, named source and qa: ","Text", Evaluatable->False ]
,Cell["
 alice = { source, qa }", "Input"]
,Cell["\n Bob only gets a single qubit: ","Text", Evaluatable->False ]
,Cell["
 bob = { qb }", "Input"]
,Cell["\n The physical system comprises Alice and Bob: ","Text", Evaluatable->False ]
,Cell["
 setSystem[ { alice, bob } ]", "Input"]
,Cell["\n All modes are indeed qubits (fermionic modes are 2-dimensional!): ","Text", Evaluatable->False ]
,Cell["
 setModeType[ #1, fermionic ] & /@ Flatten[ system ]", "Input"]
,Cell["\n Alice and Bob share one ebit of entanglement, the resource: ","Text", Evaluatable->False ]
,Cell["
 resource = singlet[ qa, qb ]", "Input"]
,Cell["\n Alice has another qubit, which is in some arbitrary and\n generally mixed state (which she wants to teleport to Bob): ","Text", Evaluatable->False ]
,Cell["
 input = generalState[ source, \"a\" ]", "Input"]
,Cell["\n Our total initial state is the product of input and resource: ","Text", Evaluatable->False ]
,Cell["
 initial = input ** resource", "Input"]
,Cell["\n The teleporation protocol starts by alice performing a bell\n measurement. We define the observable to be measured: the\n following operator's eigenstates are the 4 bell states, with\n the numbers 1..4 being the eigenvalues: ","Text", Evaluatable->False ]
,Cell["
 bellOperator = Plus @@ (\n #1 * bellState[ source, qa, #1 ] & /@ { 1, 2, 3, 4 }
\n );
", "Input"]
,Cell["\n In case you are in doubt: it is indeed an observable: ","Text", Evaluatable->False ]
,Cell["
 hermitianQ[ bellOperator ]", "Input"]
,Cell["\n Now we perform the measurement, storing the result under the\n label \"pbell\": ","Text", Evaluatable->False ]
,Cell["
 intermediate = initial ~measure~ { bellOperator, pbell };", "Input"]
,Cell["\n To complete the teleportation, the protocol now calls for Bob\n to perform one of the following maps (they are all unitary): ","Text", Evaluatable->False ]
,Cell["
 (\n   map1 = map[ { identityMatrix[ qb ] } ];
\n ; map2 = map [ { 2 * matrix[ op[ jz, qb ] ] } ]
\n ; map3 = map[ { notGate[ qb ] } ]
\n ; map4 = map[ { 2 * matrix[ op[ jz, qb ] ] ** notGate[ qb ] } ]
\n );
", "Input"]
,Cell["\n Which of these maps Bob has to perform depends on the result of\n Alice's Bell measurement, so Bob's map is a conditional map: ","Text", Evaluatable->False ]
,Cell["
 bobsMap = condmap[ {\n { map1, pbell == 1}
\n , { map2, pbell == 2}
\n , { map3, pbell == 3}
\n , { map4, pbell == 4}
\n } ]
", "Input"]
,Cell["\n We just have to apply it: ","Text", Evaluatable->False ]
,Cell["
 final = bobsMap ** intermediate;", "Input"]
,Cell["\n If we look at Bob's qubit, discarding Alice's qubits as well\n well as the classical knowledge, we find that the teleportation\n was indeed successful, with probability p=1.0: ","Text", Evaluatable->False ]
,Cell["
 final ~contract~ Not[bob] ~forget~ pbell", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n We are finished now but you may find it interesting to take a\n closer look at what was going on: ","Text", Evaluatable->False ]
,Cell["\n This is the observable of the Bell measurement: ","Text", Evaluatable->False ]
,Cell["
 bellOperator", "Input"]
,Cell["
 bellOperator // eigenvalues", "Input"]
,Cell["\n The intermediate state which was created by Alice's Bell measurement\n contains 4 different quantum states, one per measurement outcome.\n The quantum states are deliberately _not_ normalized; rather, their\n trace represents the probability of getting the corresponding result.\n Here, all 4 outcomes are equally likely: ","Text", Evaluatable->False ]
,Cell["
 intermediate", "Input"]
,Cell["\n If we just look at what Bob holds in this intermediate state,\n it becomes clear that the teleportation is already almost complete\n up to some simple transformation which depends on the outcome of\n the Bell measurement: ","Text", Evaluatable->False ]
,Cell["
 intermediate ~contract~ Not[bob]", "Input"]
,Cell["\n Knowing the result of the Bell measurement is crucial;\n if we forget it, Bob is left with a perfect mixture: ","Text", Evaluatable->False ]
,Cell["
 intermediate ~contract~ Not[bob] ~forget~ pbell", "Input"]
,Cell["\n The result of the measurement is crucial for factorization\n of Bob's qubit: ","Text", Evaluatable->False ]
,Cell["
 intermediate ~factorizes~ qb", "Input"]
,Cell["\n Without the knowledge, we get a lengthy expression but it\n is easy to verify that it cannot be fulfilled: ","Text", Evaluatable->False ]
,Cell["
 intermediate ~forget~ pbell ~factorizes~ qb // FullSimplify", "Input"]
,Cell["\n The full final state looks like this: ","Text", Evaluatable->False ]
,Cell["
 final", "Input"]
,Cell["\n Bob's quantum state is now already the correct one, independent\n of the outcome of the Bell measurement: ","Text", Evaluatable->False ]
,Cell["
 final ~contract~ Not[bob]", "Input"]
,Cell["\n Alice OTOH still holds one of the four Bell states:\n (here we have to force Mathematica to make use of the\n fact that a11 + a22 == 1) ","Text", Evaluatable->False ]
,Cell["
 final ~contract~ Not[alice] //. (a11->1-a22)", "Input"]

}]