Notebook[{
Cell[
"
This file was created automatically by vma2nb from an .vma file.
It should be suitable to be loaded by Mathematica's GUI front
end but you probably don't want to edit it; edit the corresponding
.vma source file instead!
"
, "Text", Evaluatable->False
]

,Cell["
 Print[ \"\n tutorial.qmatrix.vma: tutorial and selftest of the qmatrix
\n \\n package.
\n \\n (C) Timo Felbinger (timo@felbinger.net), 1999, 2000, 2001
\n \\n last modified:  20010430.105621utc  by: timof@uranos
\n \"]
", "Input"]
(* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; *)
(* ;;  This file is free software; you can redistribute it and/or modify     ;; *)
(* ;;  it under the terms of the GNU General Public License as published by  ;; *)
(* ;;  the Free Software Foundation; either version 2, or (at your option)   ;; *)
(* ;;  any later version.                                                    ;; *)
(* ;;                                                                        ;; *)
(* ;;  This program is distributed in the hope that it will be useful,       ;; *)
(* ;;  but WITHOUT ANY WARRANTY; without even the implied warranty of        ;; *)
(* ;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         ;; *)
(* ;;  GNU General Public License for more details.                          ;; *)
(* ;;                                                                        ;; *)
(* ;;  You should have received a copy of the GNU General Public License     ;; *)
(* ;;  along with this file (in a file called LICENSE); if not, write to     ;; *)
(* ;;  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,      ;; *)
(* ;;  Boston, MA 02111-1307, USA, or look at http://www.gnu.org.            ;; *)
(* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; *)
,Cell["\n tutorial.qmatrix:\n\n This file contains some examples of what the qmatrix package can do.\n It is meant to be evaluated interactively. Make sure to evaluate all\n input lines in order, as there are dependencies.\n\n This tutorial is by no means a complete or faithful documentation;\n it should illustrate what can be done and get you started. For more\n precise information, consult the documentation of the package, or\n look at the source code (qmatrix.vma is easily readable with any text\n editor).\n\n Of course you can play with this file, modify the input or enter new\n input lines. Note that you can also get more help on most functions\n introduced in this tutorial (see below). ","Text", Evaluatable->False ]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n We start with some bookkeeping:\n Print date, time, session ID and Mathematica Version which\n last processed this file; ","Text", Evaluatable->False ]
,Cell["
 { Date[] , $SessionID, $Version }", "Input"]
,Cell["\n Here you can change the length of output lines to whatever you like\n (and whatever fits onto your terminal / printing paper):\n (This has probably no effect in the notebook front end) ","Text", Evaluatable->False ]
,Cell["
 SetOptions[ $Output, PageWidth -> 72 ];", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n Now we load the qmatrix package: ","Text", Evaluatable->False ]
,Cell["
 << qmatrix.m", "Input"]
,Cell["\n Reading package qmatrix has created many symbols in context qmatrix`.\n Remember that you always must load this package _before_ using symbols\n from it in any way, or else the symbols may be created in the wrong\n context.\n For more information on contexts in Mathematica, evaluate the\n following: ","Text", Evaluatable->False ]
,Cell["\n   ? qmatrix`context ","Text", Evaluatable->False ]
,Cell["\n You can get help on all publicly available functions and other\n symbols defined in package qmatrix: evaluate the following symbol\n to get started. This will produce a rather long output, listing all\n symbols from the package for which additional documentation exists: ","Text", Evaluatable->False ]
,Cell["\n    qmatrix`help ","Text", Evaluatable->False ]
,Cell["\n Note that most often, you don't have to specify the context qmatrix` ","Text", Evaluatable->False ]
,Cell["\n The following will read a tiny additional file to get a space\n saving output of real and complex numbers (you do not need this\n if you don't like it): ","Text", Evaluatable->False ]
,Cell["
 << i.m", "Input"]
,Cell["\n ==================================================================== ","Text", Evaluatable->False ]
,Cell["\n Before doing anything else with the qmatrix package, you have to\n define your physical system. ","Text", Evaluatable->False ]
,Cell["\n The first thing to do is to introduce the modes we are playing with;\n here, \"modes\" are the smallest factors, or subsystems, of the total\n Hilbert space (typical example: a qubit).\n\n Of course, our modes feature the usual characters; each one of them\n has a number of subsystems at her disposal: ","Text", Evaluatable->False ]
,Cell["
 alice = { { qa1, qa2 }, atom, qutrita }", "Input"]
,Cell["
 bob = { { qb1, qb2 }, { blob, spin }, qutritb }", "Input"]
,Cell["
 eve = { { qe1, qe2, qe3, qe4 }, evilDevice, qutrite }", "Input"]
,Cell["\n These names we chose for modes (here eg qa1, qa2, ...) are completely\n arbitrary but you should never assign anything to them; they are just\n labels and nothing else.\n Also, you may choose use arbitrary names for subsystems (here eg alice),\n but you should never assign anything to them but a (possibly nested)\n list of modes. ","Text", Evaluatable->False ]
,Cell["\n Now we can define the total system to work with: ","Text", Evaluatable->False ]
,Cell["
 setSystem[ alice, bob, eve ]", "Input"]
,Cell["\n Mathematica has now learned the structure of the system we want to\n work with; the next step is to define what Hilbert spaces our \"modes\"\n do actually represent.\n This is done by calling setModeType for every mode: ","Text", Evaluatable->False ]
,Cell["\n We have lots of qubits... ","Text", Evaluatable->False ]
,Cell["
 ( setModeType[ #1, fermionic ] ) & /@ { qa1, qa2, qb1, qb2, qe1, qe2, qe3, qe4 }", "Input"]
,Cell["\n ... some qutrits... ","Text", Evaluatable->False ]
,Cell["
 ( setModeType[ #1, { discrete, 3 } ] ) & /@ { qutrita, qutritb, qutrite }", "Input"]
,Cell["\n ... a three-level atom ... ","Text", Evaluatable->False ]
,Cell["
 setModeType[ atom, { discrete, 3 } ]", "Input"]
,Cell["\n ... a bosonic mode with a maximum occupation of 3 ... ","Text", Evaluatable->False ]
,Cell["
 setModeType[ blob, { bosonic, 4 } ]", "Input"]
,Cell["\n ... a spin-1/2 particle ... ","Text", Evaluatable->False ]
,Cell["
 setModeType[ spin, { fixedAngularMomentum, 2 } ]", "Input"]
,Cell["\n ... and some evil device which has 7 orthogonal basis states: ","Text", Evaluatable->False ]
,Cell["
 setModeType[ evilDevice, { discrete, 7 } ]", "Input"]
,Cell["\n Of course, you can change these settings at any time by calling setSystem\n and setModeType again. ","Text", Evaluatable->False ]
,Cell["\n We can check whether the types and dimensions are assigned\n correctly, by asking for mode type and dimension of modes: ","Text", Evaluatable->False ]
,Cell["
 { modeType[#1], dimension[#1] } & /@ Flatten[ system ] // MatrixForm", "Input"]
,Cell["\n You can also ask for the dimension of composite systems: ","Text", Evaluatable->False ]
,Cell["
 dimension[ { qa1, qa2 } ]", "Input"]
,Cell["
 dimension[ alice ]", "Input"]
,Cell["
 dimension[ system ]", "Input"]
,Cell["\n ==================================================================== ","Text", Evaluatable->False ]
,Cell["\n Now we can define states and operators: ","Text", Evaluatable->False ]
,Cell["\n Example: Alice and Bob share an ebit, stored in their qubits qa1 and\n qb1. Because this is a very common case, there is a special function\n singlet[] which returns this state: ","Text", Evaluatable->False ]
,Cell["
 resource = singlet[ qa1, qb1 ]", "Input"]
,Cell["\n You will recognize the density matrix of an EPR pair. ","Text", Evaluatable->False ]
,Cell["\n Note that this still contains (in the last line) the full information\n on the space this object is living in. We can also see that Alice's\n index is the first, and thus the slowly changing one, while Bob's\n index the second, and thus fast changing one (ie, we use C index\n convention). ","Text", Evaluatable->False ]
,Cell["\n The above output is the default output format, but internally,\n resource is an object of the form form matrix[c,m]; by asking for\n the input form, we can get a more detailed (but less pretty) output: ","Text", Evaluatable->False ]
,Cell["
 resource // InputForm", "Input"]
,Cell["\n TeXForm is also defined for objects of the form matrix[c,m]; it produced\n the TeX code for a matrix but leaves out the space information: ","Text", Evaluatable->False ]
,Cell["
 resource // TeXForm", "Input"]
,Cell["\n Generally, any object of the form matrix[ c, m ] represents a matrix\n in the computational basis, where c contains the coordinates and the\n list m lists the modes the matrix lives in.\n\n Coordinates are stored in a nested list, with every level of nesting\n corresponding to one mode in the mode list.\n This form is very useful for further manipulation by Mathematica but\n may look somewhat confusing to humans at first glance. ","Text", Evaluatable->False ]
,Cell["\n You can use \"flatten\" to get a more familiar depth-2 representation;\n again, only the InputForm will reveal all details: ","Text", Evaluatable->False ]
,Cell["
 resource //flatten //InputForm", "Input"]
,Cell["\n This is now a 2-dimensional \"flat matrix\", denoted by fm instead of\n matrix. The default output format of any matrix performs this conversion\n to human-readable form internally, so you will usually only see the\n 2-dimensional form unless you ask excplicitely for the InputForm. ","Text", Evaluatable->False ]
,Cell["\n Now we can start doing some simple computations: ","Text", Evaluatable->False ]
,Cell["\n resource is a pure state which can easily be verified by multiplying\n it with itself ","Text", Evaluatable->False ]
,Cell["
 resource ** resource", "Input"]
,Cell["\n The operator ** can be used to form the \"Dirac product\" of matrices,\n which is the outer product (aka tensor product) except that whenever\n a certain mode occurs as a bra in the left factor and as a ket in\n the right factor, this is understood as a \"braket\" (as in Dirac's\n notation) and the inner product of these modes is computed. ","Text", Evaluatable->False ]
,Cell["\n In cases where purity is not obvious, the following is an\n easy way to check: ","Text", Evaluatable->False ]
,Cell["
 resource ** resource == resource", "Input"]
,Cell["\n Yet another way to check for purity is to compute the entropy: ","Text", Evaluatable->False ]
,Cell["
 neumannEntropy[ resource ]", "Input"]
,Cell["\n As the singlet state is pure, we could just as well start from a\n state vector instead; it's a column vector: ","Text", Evaluatable->False ]
,Cell["
 resourceKet = singletKet[ qa1, qb1 ]", "Input"]
,Cell["\n pretty-printing will, among other simplifications, replace numerical\n expressions like roots by approximations in decimal notation. This\n happens only to get a consistent and space-saving output format, the\n internal representation is of course exact: ","Text", Evaluatable->False ]
,Cell["
 resourceKet // InputForm", "Input"]
,Cell["\n There are functions ketQ and braQ to verify that given objects are\n indeed kets or bras (these functions are mainly for internal use): ","Text", Evaluatable->False ]
,Cell["
 resourceKet // braQ", "Input"]
,Cell["
 resourceKet // ketQ", "Input"]
,Cell["\n There is a function hc[] which computes the hermitian conjugate of\n any matrix. Applying it to a ket will return a bra, which prints as\n a line vector: ","Text", Evaluatable->False ]
,Cell["
 resourceKet // hc", "Input"]
,Cell["\n This may look incorrect at first glance (as the minus sign has changed\n position) but note that the order of modes in the mode list was reversed.\n What you got is indeed one correct representation of the corresponding bra.\n You can control the order of modes if you like (be patient, we will come to\n this later). ","Text", Evaluatable->False ]
,Cell["\n The tensor product with its own hermitian conjugate yields again the\n density matrix: ","Text", Evaluatable->False ]
,Cell["
 resourceKet ** hc[ resourceKet ]", "Input"]
,Cell["\n The order of factors in the Dirac product ** is of course important;\n if you exchange the factors, the inner product will be computed: ","Text", Evaluatable->False ]
,Cell["
 hc[ resourceKet ] ** resourceKet // InputForm", "Input"]
,Cell["\n This is still considered a 1x1 matrix (in a 1-dimensional space); if\n you want a c-number, compute the trace: ","Text", Evaluatable->False ]
,Cell["
 hc[ resourceKet ] ** resourceKet // trace", "Input"]
,Cell["\n Besides the ** operator, you can also compute scalar multiples of\n matrices... ","Text", Evaluatable->False ]
,Cell["
 42 resource", "Input"]
,Cell["\n ... as well as sums and differences of matrices: ","Text", Evaluatable->False ]
,Cell["
 resource + hc[resource]", "Input"]
,Cell["
 resource - hc[resource]", "Input"]
,Cell["\n Of course, you cannot add kets and bras: ","Text", Evaluatable->False ]
,Cell["
 resourceKet + hc[ resourceKet ]", "Input"]
,Cell["\n ==================================================================== ","Text", Evaluatable->False ]
,Cell["\n You may be interested in other ways to create matrices: ","Text", Evaluatable->False ]
,Cell["\n Many well-known operators can be represented by objects of the\n form op[ <type>, <mode>, ... ]; they can be turned into their\n matrix representation, for example... ","Text", Evaluatable->False ]
,Cell["\n ...the bosonic creation operator ad... ","Text", Evaluatable->False ]
,Cell["
 matrix[ op[ ad, blob ] ]", "Input"]
,Cell["\n ...or matrices representing the operators a and ad (considered as\n ladder operators, as in the case of a bosonic mode) of our spin: ","Text", Evaluatable->False ]
,Cell["
 matrix[ op[ a, spin ] ]", "Input"]
,Cell["
 matrix[ op[ ad, spin ] ]", "Input"]
,Cell["\n The spin components are available, too: ","Text", Evaluatable->False ]
,Cell["
 sx = matrix[ op[ jx, spin ] ]", "Input"]
,Cell["
 sy = matrix[ op[ jy, spin ] ]", "Input"]
,Cell["
 sz = matrix[ op[ jz, spin ] ]", "Input"]
,Cell["\n (Note that we number basis states in order of increasing z-component,\n which is different from the convention used in many texts) ","Text", Evaluatable->False ]
,Cell["\n These matrices fullfill the usual commutation relation of a spin\n (and there is a special function to compute commutators): ","Text", Evaluatable->False ]
,Cell["
 sx ~commutator~ sy == I sz", "Input"]
,Cell["\n The spin matrices can be used to compute arbitrary rotations; we make\n use of the exponential function for matrices: ","Text", Evaluatable->False ]
,Cell["
 rot = exp[ 2*Pi*I/3 * sx ]", "Input"]
,Cell["
 rot ** rot ** rot == identityMatrix[ spin ]", "Input"]
,Cell["\n identityMatrix can give you the identity matrix for any\n subsystem: ","Text", Evaluatable->False ]
,Cell["
 identityMatrix[ qa1, blob ]", "Input"]
,Cell["\n The j-operators are defined for non-spin modes, too: ","Text", Evaluatable->False ]
,Cell["
 matrix[ op[ jy, qa1 ] ]", "Input"]
,Cell["\n Here the qubit qa1 was treated like a spin-1/2 system ","Text", Evaluatable->False ]
,Cell["\n For non-spins, jPlus acts differently from ad: ","Text", Evaluatable->False ]
,Cell["
 matrix[ op[ ad, blob ] ] == matrix[ op[ jPlus, blob ] ]", "Input"]
,Cell["\n This abstract operator notation op[ name, mode ] could in some later\n version of this package (sorry, not in the current one) or by an\n additional package be used to allow some abstract operations, eg by\n teaching Mathematica commutation relations. ","Text", Evaluatable->False ]
,Cell["\n There are also functions which give you frequently used matrices\n without involving an abstract operator first. Many of them are old\n friends from quantum information theory, for example ","Text", Evaluatable->False ]
,Cell["\n ...the cnot gate... ","Text", Evaluatable->False ]
,Cell["
 cnotGate[ qa1, qb1 ]", "Input"]
,Cell["\n ...a werner state... ","Text", Evaluatable->False ]
,Cell["
 wernerState[ 0.32, qa1, qa2 ]", "Input"]
,Cell["\n ...or any of the four bell basis states: ","Text", Evaluatable->False ]
,Cell["
 bellStateKet[ qa1, qa2, #1 ] & /@ { 1, 2, 3, 4 }", "Input"]
,Cell["\n We can also get the operator of the discrete fourier transform: ","Text", Evaluatable->False ]
,Cell["
 fourierTransform[ qa1 ]", "Input"]
,Cell["\n It works for higher dimensions, too: ","Text", Evaluatable->False ]
,Cell["
 fourierTransform[ evilDevice ]", "Input"]
,Cell["\n The quite long delay when evaluating the above input was caused by\n the pretty-printing, not by the fourierTransform itself.\n Using a numerical approximation, where approbriate, is much faster: ","Text", Evaluatable->False ]
,Cell["
 fte = fourierTransform[ evilDevice ] // N", "Input"]
,Cell["\n The resulting operator is indeed unitary: ","Text", Evaluatable->False ]
,Cell["
 fte // ( #1 ** hc[#1] == identityMatrix[ evilDevice ] ) &", "Input"]
,Cell["\n You can of course create arbitrary matrices explicitely: ","Text", Evaluatable->False ]
,Cell["
 mystate = matrix[ { { 1.7 ,6.2I - 8.1}, {-6.3I, 4I-5} }, {ket[qa1],bra[qa1] } ]", "Input"]
,Cell["\n The coordinates need not be numbers; there is a special function to\n create a mixed state with general unknown coordinates.\n We define one for qubit qa1, with coordinates named \"a\": ","Text", Evaluatable->False ]
,Cell["
 sa2 = generalState[ qa2, \"a\" ]", "Input"]
,Cell["\n We create a similar state for qubit qb2, naming the coordinates \"b\": ","Text", Evaluatable->False ]
,Cell["
 sb2 = generalState[ qb2, \"b\" ]", "Input"]
,Cell["\n Defining these states had an important side effect: As these are\n assumed to be normalized density matrices, Mma learned important\n rules: ","Text", Evaluatable->False ]
,Cell["
 a11 + a22", "Input"]
,Cell["\n This allows Mathematica eg to compute traces: ","Text", Evaluatable->False ]
,Cell["
 sa2 ** sb2 // trace", "Input"]
,Cell["\n Keep this in mind: these rules will remain in Mathematica's memory\n even if you use the variables a11, a22 for a different purpose! ","Text", Evaluatable->False ]
,Cell["\n These rules do not always work perfectly: oftentimes, Mathematica\n will ; see that an expression can be simplified, and you may have to\n help it, eg: ","Text", Evaluatable->False ]
,Cell["
 -a11 - a22", "Input"]
,Cell["
 ( -a11 - a22 ) //. (a11->1-a22)", "Input"]
,Cell["\n You _can_, if you like, add more rules, like\n    a11 := 1 - a22\n to Mathematica's rule set, but this is probably not a good idea:\n most likely you don't want this replacement to be performed\n everywhere! ","Text", Evaluatable->False ]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n There are many functions to check properties of objects: ","Text", Evaluatable->False ]
,Cell["
 matrixQ[ resource ]", "Input"]
,Cell["
 flatMatrixQ[ resource ]", "Input"]
,Cell["
 flatMatrixQ[ resource // flatten ]", "Input"]
,Cell["
 qubitQ[ qa1 ]", "Input"]
,Cell["
 qubitQ[ atom ]", "Input"]
,Cell["\n There are also functions to find modes and subsystems in the\n system: ","Text", Evaluatable->False ]
,Cell["
 system", "Input"]
,Cell["\n This is a nested list representing a \"natural\" structure;\n we can ask for subsystems of the total system containing \"blob\"\n at different levels: ","Text", Evaluatable->False ]
,Cell["
 subSystem[ blob, 1 ]", "Input"]
,Cell["
 subSystem[ blob, 2 ]", "Input"]
,Cell["
 subSystem[ blob, 3 ]", "Input"]
,Cell["\n however we can also ask for subsystems: ","Text", Evaluatable->False ]
,Cell["
 subSystem[ alice, 1 ]", "Input"]
,Cell["
 subSystem[ { blob, spin } , 1 ]", "Input"]
,Cell["\n The following will return an error as { blob, foo } is not a\n subsystem: ","Text", Evaluatable->False ]
,Cell["
 subSystem[ { blob, foo } , 1 ]", "Input"]
,Cell["\n The following may be unexpected: as for all matrices,\n    dimension[ resource ]\n is really the dimension of the tensor space this operator lives in,\n not - as one might expect - the dimension of the space it acts on! ","Text", Evaluatable->False ]
,Cell["
 dimension[ resource ]", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n Now we can start to create some larger objects, starting from smaller\n ones: ","Text", Evaluatable->False ]
,Cell["
 resource", "Input"]
,Cell["\n ... and some arbitrary state sa2: ","Text", Evaluatable->False ]
,Cell["
 sa2", "Input"]
,Cell["\n We can define a state (let's call it initial) to be the product\n of both:\n Note that Mathematica's operator precedende rules require the\n parentheses in the following input: ","Text", Evaluatable->False ]
,Cell["
 initial = resource ** sa2", "Input"]
,Cell["\n The function factorizes[] can be used to check whether a given matrix\n can be factorized in a factor living in a specified space, and a\n factor living outside this space.\n For example, we can easily verify that initial is indeed a product of\n the qubit qa2 and something else... ","Text", Evaluatable->False ]
,Cell["
 initial ~factorizes~ qa2", "Input"]
,Cell["\n ...while of course it is not a product of alice and bob:\n Unfortunately Mma isn't smart enough to know that not all coefficients\n can vanish simultaneously, but for us humans it is easy to see that\n the resulting lengthy condition cannot hold: ","Text", Evaluatable->False ]
,Cell["
 initial ~factorizes~ bob", "Input"]
,Cell["\n The function contract can be used to contract matrices over specified\n spaces (aka partial trace). ","Text", Evaluatable->False ]
,Cell["\n We can check that after tracing out alice, bob holds a perfect\n mixture: ","Text", Evaluatable->False ]
,Cell["
 initial ~contract~ alice", "Input"]
,Cell["\n OTOH Alice holds something slightly more complicated: ","Text", Evaluatable->False ]
,Cell["
 initial ~contract~ bob", "Input"]
,Cell["\n It is often more convenient to specify the modes we want to keep,\n instead of the modes to contract. A special notation allows this:\n the following is another way to get Alice's state (in fact, it is\n even safer as it will still be correct after Eve enters the game): ","Text", Evaluatable->False ]
,Cell["
 initial ~contract~ Not[alice]", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n To illustrate the use of ** note that it will compute\n the inner product when appropriate: ","Text", Evaluatable->False ]
,Cell["
 sa2 ** sa2", "Input"]
,Cell["\n ... and the outer product when no inner product can be\n computed: ","Text", Evaluatable->False ]
,Cell["
 sa2 ** sb2", "Input"]
,Cell["\n You can always enforce the computation of the outer product\n (though this is rarely a good idea): ","Text", Evaluatable->False ]
,Cell["
 outer$sa2sa2 = outerMatrixProduct[ sa2, sa2 ]", "Input"]
,Cell["\n This matrix wasn't pretty-printed because it is not a well-formed\n matrix, because it contains brakets (think about it: how should this\n thing look after pretty-printing? I don't have any idea...): ","Text", Evaluatable->False ]
,Cell["
 properMatrixQ[ outer$sa2sa2 ]", "Input"]
,Cell["\n You can of course contract this matrix to make it proper: ","Text", Evaluatable->False ]
,Cell["
 outer$sa2sa2 // contractAll // properMatrixQ", "Input"]
,Cell["\n The result is exactly what you would get by computing the Dirac product\n in the first place: ","Text", Evaluatable->False ]
,Cell["
 ( outer$sa2sa2 // contractAll ) == sa2 ** sa2", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n Let's try some other frequently used operators:\n state swappers, for example, can be created for any pair of modes\n of equal dimension: ","Text", Evaluatable->False ]
,Cell["
 matrix[ op[ basis, qa1, 1, 2 ] ]", "Input"]
,Cell["
 stateSwapper[ qa1, qa2 ]", "Input"]
,Cell["\n We can multiply such matrices to get some big one: ","Text", Evaluatable->False ]
,Cell["
 stateSwapper[ qa1, qa2 ] ** stateSwapper[ qb1, qb2 ]", "Input"]
,Cell["\n And we can get even larger: remove the ; at the end of the following\n input to get a nice matrix with self-similar features: ","Text", Evaluatable->False ]
,Cell["
 bigswap =\n stateSwapper[ qa1, qa2 ] ** stateSwapper[ qb1, qb2 ] ** stateSwapper[ qe1, qe2 ];
", "Input"]
,Cell["\n And even larger than that; remove the ; at the end of the following\n input if you want to see the resulting 256x256 matrix:\n *** WARNING: This may crash the GUI front end! ","Text", Evaluatable->False ]
,Cell["
 hugeswap =\n stateSwapper[ qa1, qa2 ] ** stateSwapper[ qb1, qb2 ] **
\n stateSwapper[ qe1, qe2 ] ** stateSwapper[ qe3, qe4 ] ;
", "Input"]
,Cell["
 dimension[ hugeswap ]", "Input"]
,Cell["\n We can recover Alice's state swapper, times a constant,\n by tracing over bob and eve: ","Text", Evaluatable->False ]
,Cell["
 hugeswap ~contract~ Not[alice]", "Input"]
,Cell["\n Swapping twice gives the identity (as it should): ","Text", Evaluatable->False ]
,Cell["
 bigswap ** bigswap == identityMatrix[ qa1, qa2, qb1, qb2, qe1, qe2 ]", "Input"]
,Cell["\n If we swap Alice's qubits again, the resulting operator should\n act like the identity (modulo normalization) on her space: ","Text", Evaluatable->False ]
,Cell["
 hugeswapEveBob = hugeswap ** stateSwapper[ qa1, qa2 ];", "Input"]
,Cell["\n This is indeed true: ","Text", Evaluatable->False ]
,Cell["
 hugeswapEveBob ~contract~ Not[alice]", "Input"]
,Cell["\n We can define an operator which swaps on Alice's space and does\n nothing to Bob's space in a somewhat creative way: ","Text", Evaluatable->False ]
,Cell["
 swapAlice =\n stateSwapper[ qa1, qb1 ] ** stateSwapper[ qa2, qb2 ] **
\n stateSwapper[ qb1, qb2 ] **
\n stateSwapper[ qa1, qb1 ] ** stateSwapper[ qa2, qb2 ]
", "Input"]
,Cell["
 swapAlice ~contract~ Not[alice]", "Input"]
,Cell["\n Now choose a slightly more interesting initial state, with eve\n involved: ","Text", Evaluatable->False ]
,Cell["
 initial = sa2 ** singlet[ qa1, qe1 ] ** sb2 ** singlet[ qb1, qe2 ];", "Input"]
,Cell["\n Note how the reduced states of qubits qa1 and qa2 look like: ","Text", Evaluatable->False ]
,Cell["
 initial ~contract~ Not[ qa1 ]", "Input"]
,Cell["
 initial ~contract~ Not[ qa2 ]", "Input"]
,Cell["\n In the swapped state, we find these interchanged, as they should: ","Text", Evaluatable->False ]
,Cell["
 final = swapAlice ** initial ** hc[ swapAlice ];", "Input"]
,Cell["
 final ~contract~ Not[ qa1 ]", "Input"]
,Cell["
 final ~contract~ Not[ qa2 ]", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n Now let's try to get some matrices which are not so sparse; for\n example a 5x5 matrix full of random numbers. It is most likely\n neither symmetric nor positive: ","Text", Evaluatable->False ]
,Cell["
 skewH = matrix[\n Outer[
\n   Random[] + I Random[] &
\n , spin // dimension // Range // Sequence @@ {#,#} &
\n ]
\n , { ket[spin], bra[spin] }
\n ]
", "Input"]
,Cell["\n We can still compute exp of log of this matrix, and by definition we\n will get back the original (within numerical precision): ","Text", Evaluatable->False ]
,Cell["
 exp[ log[ skewH ] ] == skewH", "Input"]
,Cell["\n Keep in mind that comparison of matrices via == always allows for\n small numerical errors. There is a predefined epsilon which you can\n change at any time if you want: ","Text", Evaluatable->False ]
,Cell["
 qmatrix`Private`epsilon", "Input"]
,Cell["\n log of exp of this matrix will often also reproduce the original but\n this need not always work out because log is not a unique function,\n so the following may legitimately return either True or False: ","Text", Evaluatable->False ]
,Cell["
 log[ exp[ skewH ] ] == skewH", "Input"]
,Cell["\n We now define a hermitian operator: ","Text", Evaluatable->False ]
,Cell["
 H = skewH + hc[skewH]", "Input"]
,Cell["\n We can find the eigenvectors and eigenvalues of H; we can\n also request that the eigenvectors be normalized: ","Text", Evaluatable->False ]
,Cell["
 normalizedEigensystem[ H ] // ColumnForm", "Input"]
,Cell["\n Let's compute a unitary matrix: ","Text", Evaluatable->False ]
,Cell["
 U = exp[ I * H ]", "Input"]
,Cell["\n Within numerical precision, U is, indeed, unitary: ","Text", Evaluatable->False ]
,Cell["
 U ** hc[U] == identityMatrix[ spin ]", "Input"]
,Cell["\n The function inverse will compute the inverse of a given matrix\n (which must of course be an operator matrix: lonly kets or bras\n  have no inverse).\n For unitary U, inverse[U] should be the same as hc[U]: ","Text", Evaluatable->False ]
,Cell["
 hc[U] == inverse[U]", "Input"]
,Cell["\n For hermitian H, both exp[log[#]]& and log[exp[#]]& are identity maps: ","Text", Evaluatable->False ]
,Cell["
 exp[ log[ H ] ] == H", "Input"]
,Cell["
 log[ exp[ H ] ] == H", "Input"]
,Cell["\n For the anti-Hermitian I*H however, this is still true for exp[log[#]]&... ","Text", Evaluatable->False ]
,Cell["
 exp[ log[ I*H] ] == I*H", "Input"]
,Cell["\n ...but in general not for log[exp[#]]& (of course, you might be\n lucky and get a True): ","Text", Evaluatable->False ]
,Cell["
 log[ exp[ I*H] ] == I*H", "Input"]
,Cell["\n square roots of matrices are defined, too: ","Text", Evaluatable->False ]
,Cell["
 sqrt[ skewH ]", "Input"]
,Cell["
 sqrt[skewH] ** sqrt[skewH] == skewH", "Input"]
,Cell["\n sqrt[], like log[], is not a unique function, so the following will\n often return False (again, you may of course be lucky): ","Text", Evaluatable->False ]
,Cell["
 skewH == sqrt[ skewH ** skewH ]", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n Much larger matrices can still be handled: ","Text", Evaluatable->False ]
,Cell["
 randomEvil = matrix[\n Outer[
\n   Random[] + I Random[] &
\n , evilDevice // dimension // Range // Sequence @@ {#,#} &
\n ]
\n , { ket[evilDevice], bra[evilDevice] }
\n ]
", "Input"]
,Cell["
 bigEvil = randomEvil ** skewH", "Input"]
,Cell["\n We can still verify that this matrix does, indeed, factorize: ","Text", Evaluatable->False ]
,Cell["
 bigEvil ~factorizes~ eve", "Input"]
,Cell["\n Here neglecting some numerical error is clearly crucial or the test\n would most certainly fail. ","Text", Evaluatable->False ]
,Cell["\n log and exp still work pretty well: ","Text", Evaluatable->False ]
,Cell["
 exp[ log[ bigEvil ] ] == bigEvil", "Input"]
,Cell["\n And even for the following \"monster\" we can still compute the eigenvalues: ","Text", Evaluatable->False ]
,Cell["
 monster = bigEvil ** sqrt[ stateSwapper[ qa1, qa2 ] ];", "Input"]
,Cell["
 dimension[ monster ]", "Input"]
,Cell["
 eigenvalues[ monster ]", "Input"]
,Cell["\n You may sometimes need truly random state vectors; the function\n randomKet will return such vectors for arbitrary Hilbert spaces: ","Text", Evaluatable->False ]
,Cell["
 kets = ( randomKet[ qa1, qa2 ] & /@ Range[5] )", "Input"]
,Cell["\n gramSchmidt can be used to orthogonalize any set of vectors; of\n course, the number of orthogonal kets returned may be smaller than\n the argument set: ","Text", Evaluatable->False ]
,Cell["
 ketsON = gramSchmidt[ kets ]", "Input"]
,Cell["\n these kets are indeed an orthogonal basis, except for some numerical\n garbage which we have to cast away: ","Text", Evaluatable->False ]
,Cell["
\n Plus @@ ( projector /@ ketsON ) //.  { x_Complex /; Abs[x] < 10^-12 -> 0 }
", "Input"]
,Cell["\n Another important concept is the Schmidt decomposition which we\n can compute for arbitrary kets: ","Text", Evaluatable->False ]
,Cell["
 v = randomKet[ qa1, qa2, qb1 ]", "Input"]
,Cell["
 schmidt = v ~schmidtDecomposition~ qb1", "Input"]
,Cell["\n the desult is a list of triples; every triple contains two eigenkets\n and the corresponding eigenvalue: ","Text", Evaluatable->False ]
,Cell["
 schmidt[[1]]", "Input"]
,Cell["\n at first glance, Schmidt decomposition produced 4 pieces: ","Text", Evaluatable->False ]
,Cell["
 Length[ schmidt ]", "Input"]
,Cell["\n however, most eigenvalues are essentially zero: ","Text", Evaluatable->False ]
,Cell["
 #3 & @@ #1 & /@ schmidt // ColumnForm", "Input"]
,Cell["\n function schmidtRank takes care of this and returns the number\n of non-zero eigenvalues: ","Text", Evaluatable->False ]
,Cell["
 v ~schmidtRank~ qb1", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n\n *** permutations: ***\n ","Text", Evaluatable->False ]
,Cell["\n Some auxiliary functions dealing with permutations are defined: ","Text", Evaluatable->False ]
,Cell["
 permute[ { foo, bar, baz, smurf }, {3,1,2,4} ]", "Input"]
,Cell["\n The second argument must be a permutation; if it is not, you will\n get an error message: ","Text", Evaluatable->False ]
,Cell["
 permute[ { foo, bar, baz, smurf }, {3,2,2,4} ]", "Input"]
,Cell["\n we can turn a permutation into a unitary operator, which permutes\n basis states of the computational basis: ","Text", Evaluatable->False ]
,Cell["
 pm = matrix[ op[ permutation, blob, {3,1,2,4} ] ]", "Input"]
,Cell["\n permutations can be inverted: ","Text", Evaluatable->False ]
,Cell["
 pminv = matrix[ op[ permutation, blob, inversePermutation[{3,1,2,4}] ] ]", "Input"]
,Cell["\n This gives indeed the inverse operator: ","Text", Evaluatable->False ]
,Cell["
 pm ** pminv", "Input"]
,Cell["\n permutation pm is a group element of order 3: ","Text", Evaluatable->False ]
,Cell["
 #1 == identityMatrix[blob] & /@ NestList[ pm ** #1 &, pm, 12 ]", "Input"]
,Cell["
 pm ** pm == pminv", "Input"]
,Cell["\n the function signedPermutations returns a list of all permutations\n of a given list and also gives the signs of these permutations: ","Text", Evaluatable->False ]
,Cell["
 signedPermutations[ { foo, bar, baz } ] // MatrixForm", "Input"]
,Cell["\n We can now easily sum over all permutations and add the\n products of the unitary operators for several spaces\n to get nice matrix: ","Text", Evaluatable->False ]
,Cell["
 psum = Plus @@ (\n    matrix[ op[ permutation, qutrita, #1 ] ]
\n ** matrix[ op[ permutation, qutritb, #1 ] ]
\n ** matrix[ op[ permutation, qutrite, #1 ] ]
\n   & /@ ( #1 & @@ Transpose[ signedPermutations[ Range[ dimension[qutrita] ] ] ] )
\n )
", "Input"]
,Cell["\n For eve only this operator acts like this: ","Text", Evaluatable->False ]
,Cell["
 psum ~contract~ Not[eve]", "Input"]
,Cell["\n This operator clearly cannot be expected to factorize: ","Text", Evaluatable->False ]
,Cell["
 psum ~factorizes~ bob", "Input"]
,Cell["\n psum is, however, hermitian: ","Text", Evaluatable->False ]
,Cell["
 hermitianQ[ psum ]", "Input"]
,Cell["
 neumannEntropy[ normalize[ psum ] ]", "Input"]
,Cell["\n psum is singular which will be discovered when trying to invert it: ","Text", Evaluatable->False ]
,Cell["
 inverse[ psum ];", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n For a state in a composite system, we can compute the partial\n transpose with respect to some mode. Recall our initial ebit\n resource: ","Text", Evaluatable->False ]
,Cell["
 resource", "Input"]
,Cell["\n We can compute the partial transpose with respect to a given mode: ","Text", Evaluatable->False ]
,Cell["
 resourceTA = resource ~partialTranspose~ qa1", "Input"]
,Cell["\n resource and resourceTA have the well-known properties: ","Text", Evaluatable->False ]
,Cell["
 resource // positiveQ", "Input"]
,Cell["
 resourceTA // positiveQ", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n Let's start again with some state living in several modes; we can\n reuse the previously defined symbols sa2 and sb2: ","Text", Evaluatable->False ]
,Cell["
 state = sa2 ** sb2", "Input"]
,Cell["\n Here the order of modes in the mode list is quite arbitrary; as long\n as no bra moves in front of a ket of the same mode (which would create\n a braket and render the matrix improper and with new semantics) we can\n permute modes as we like and arrive at different representations of the\n same state, provided that we transpose the coordinate list accordingly.\n The function reorderMatrix[] accomplishes this: ","Text", Evaluatable->False ]
,Cell["
 statePerm =\n state ~reorderMatrix~ { ket[qa2], ket[qb2], bra[qb2], bra[qa2] }
", "Input"]
,Cell["\n Note that pretty-printing has special rules for ordering modes:\n as it is required for 2-dimensional representations, all kets (which\n determine the row indices) come first and all bras (which determine\n the column indices) come last. Moreover, for operator matrices where\n kets and bras occur in pairs, the order of kets will be kept and the\n bras will be reordered to match the order of kets, which is what human\n readers usually expect. Internally, the new order of modes has taken\n effect: ","Text", Evaluatable->False ]
,Cell["
 statePerm // InputForm", "Input"]
,Cell["\n state and statePerm, though having quite different coordinate lists,\n are indeed semantically equivalent: ","Text", Evaluatable->False ]
,Cell["
 state - statePerm", "Input"]
,Cell["\n We just saw that the + (or -) operator for matrices can cope with\n matrices where the mode list is permuted. It is even much more\n powerful: ","Text", Evaluatable->False ]
,Cell["
 myop = stateSwapper[ qa1, qb1 ] + stateSwapper[ qb1, qe1 ]", "Input"]
,Cell["\n We have just added two operators living in different Hilbert spaces.\n This is accomplished by implicitely multiplying each operator with\n suitable identity operators so that both are operating on the least\n common multiple of both Hilbert spaces.\n\n CAVEAT: this procedure is usually ok for actual OPERATORS but not for\n STATE OPERATORS as it violates the normalization; moreover, it would\n be quite strong an assumption to assume a perfect mixture for modes\n which are added implicitely. ","Text", Evaluatable->False ]
,Cell["\n Of course, this procedure must fail for matrices which are not\n operators: ","Text", Evaluatable->False ]
,Cell["
 singletKet[ qa1, qb1 ] + singletKet[ qb1, qe1 ]", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n So far, we have a powerful set of low-level tools at hand;\n now let's look at some even more powerful high-level objects: ","Text", Evaluatable->False ]
,Cell["\n We start from a new resource: 2 ebits: ","Text", Evaluatable->False ]
,Cell["
 r0 = singlet[ qa1, qb1 ] ** singlet[ qa2, qb2 ];", "Input"]
,Cell["\n To make matrices more easily comparable we define a standard\n order of modes: ","Text", Evaluatable->False ]
,Cell["
 order = Flatten[ Transpose[ { ket[#1], bra[#1] } & /@ { qa1, qb1, qa2, qb2 } ] ]", "Input"]
,Cell["
 r0 ~reorderMatrix~ order", "Input"]
,Cell["\n To make it a bit more interesting, we mix qa1 and qa2 somewhat: ","Text", Evaluatable->False ]
,Cell["
 swaproot = sqrt[ stateSwapper[ qa1, qa2 ] ];", "Input"]
,Cell["
 r1 = ( swaproot ** r0 ** hc[swaproot] ) ~reorderMatrix~ order", "Input"]
,Cell["\n We can get the same thing by using swaproot as a \"map\": ","Text", Evaluatable->False ]
,Cell["
 map[ { swaproot } ] ** r0 == r1", "Input"]
,Cell["\n maps are of course more powerful than this and accept more than\n one (Kraus-)operator, eg: ","Text", Evaluatable->False ]
,Cell["
 (\n r3 = map[ { Sqrt[1/2] * swaproot, Sqrt[1/2] * identityMatrix[ qa1 ] } ] ** r0
\n ) ~reorderMatrix~ order
", "Input"]
,Cell["\n Usually the Kraus operators in a map will satisfy additional\n properties; these restrictions are however not enforced by this\n package. ","Text", Evaluatable->False ]
,Cell["\n We can perform measurements on matrices:\n eg we can measure the state of mode qb1, using the computational\n basis as the pointer basis: ","Text", Evaluatable->False ]
,Cell["
 qc1 = r3 ~measure~ { qb1, pointerb1 };", "Input"]
,Cell["\n The resulting object is a \"qcstate\" which combines quantum and\n classical information; pretty-printing will give you several\n matrices, corresponding to mutually exclusive classical knowledge.\n The knowledge (ie pointer values) and probabilities are listed, too: ","Text", Evaluatable->False ]
,Cell["
 qc1 ~reorderMatrix~ order", "Input"]
,Cell["\n Mode qb1 in the state before measurement does of course not factorize: ","Text", Evaluatable->False ]
,Cell["
 r3 ~factorizes~ qb1", "Input"]
,Cell["\n After the measurement, however, it does factorize, independent of the\n outcome of the measurement: ","Text", Evaluatable->False ]
,Cell["
 qc1 ~factorizes~ qb1", "Input"]
,Cell["\n qa1 still can not be factorized: ","Text", Evaluatable->False ]
,Cell["
 qc1 ~factorizes~ qa1", "Input"]
,Cell["\n These measurements are nondestructive; by tracing out everything\n else we can verify that the quantum states are indeed eigenstates of\n the measured observable: ","Text", Evaluatable->False ]
,Cell["
 qc1 ~contract~ Not[qb1]", "Input"]
,Cell["\n Note that the quantum states in a qcstate are NOT normalized;\n rather, their trace is to be interpreted as a probability of\n ending up in this state. ","Text", Evaluatable->False ]
,Cell["\n qcstates can be subjected to further measurements; instead of\n the computational basis, we can also use arbitrary observables,\n eg the x-component of a spin;\n the resulting qcstate then has 4 components: ","Text", Evaluatable->False ]
,Cell["
 qc2 = qc1 ~measure~ { matrix[ op[ jx, qa1 ] ], pointerjxa1 }", "Input"]
,Cell["\n tracing out everything but qa1, we find that these are\n indeed jx-eigenstates: ","Text", Evaluatable->False ]
,Cell["
 qc2 ~contract~ Not[qa1]", "Input"]
,Cell["\n Even after tracing out quantum information of mode qb1, the\n classical result is still present, so we just got 4 components.\n If we are not interested in it any more, we can \"forget\" the\n state of classical pointers: ","Text", Evaluatable->False ]
,Cell["
 qc2 ~contract~ Not[qa1] ~forget~ pointerb1", "Input"]
,Cell["\n If we forget the result of the jx-measurement, too, we end\n up with a perfect mixture: ","Text", Evaluatable->False ]
,Cell["
 qc2 ~contract~ Not[qa1] ~forget~ { pointerb1, pointerjxa1 }", "Input"]
,Cell["\n We can see that the classical information is crucial for the\n mode qa1 to factorize in the state qc2: ","Text", Evaluatable->False ]
,Cell["
 qc2 ~factorizes~ qa1 // ColumnForm", "Input"]
,Cell["\n If we forget the classical information, we create\n classical correlations which prevent factorization: ","Text", Evaluatable->False ]
,Cell["
 qc2 ~forget~ pointerjxa1 ~factorizes~ qa1 // ColumnForm", "Input"]
,Cell["\n Observables can have degenerate eigenvalues; as an extreme example,\n we can \"measure\" the identity operator, which gives the result 1 with\n probability 1 and thus yields absolutely no classical information: ","Text", Evaluatable->False ]
,Cell["
 r3 ~measure~ { identityMatrix[ qa1 ], pid }", "Input"]
,Cell["\n As a slightly more interesting example, we can measure the probability\n of finding qa1 in state 2 and qb2 in state 1: ","Text", Evaluatable->False ]
,Cell["
 r3 ~measure~ {\n matrix[ op[ basis, qa1, 2 ] ] ** matrix[ op[ basis, qb2, 1 ] ]
\n , somepointer
\n }
", "Input"]
,Cell["\n -------------------------------------------------------------------- ","Text", Evaluatable->False ]
,Cell["\n Many protocols ask for conditional application of maps,\n depending on classical information. There is a special\n object to implement such protocols: ","Text", Evaluatable->False ]
,Cell["\n Let's first define two simple unitary maps: ","Text", Evaluatable->False ]
,Cell["
 m1 = map[ { identityMatrix[ qa1 ] } ]", "Input"]
,Cell["
 m2 = map[ { stateSwapper[ qb1, qa1 ] } ]", "Input"]
,Cell["\n We want to apply m1 if the jx-measurement gave result -1/2,\n and m2 if we got +1/2: ","Text", Evaluatable->False ]
,Cell["
 cm1 = condmap[ {\n { m1, pointerjxa1 == -1/2 }
\n , { m2, pointerjxa1 == +1/2 }
\n } ]
", "Input"]
,Cell["\n Now apply this conditional map: ","Text", Evaluatable->False ]
,Cell["
 qc3 = cm1 ** qc2;", "Input"]
,Cell["\n We can see that for jz=+1/2, the states of qa1 and qb1\n were indeed exchanged: ","Text", Evaluatable->False ]
,Cell["
 qc3 ~contract~ Not[qb1]", "Input"]
,Cell["
 qc3 ~contract~ Not[qa1]", "Input"]
,Cell["\n If none of the conditions of a conditional map can be\n fulfilled, you will get an error: ","Text", Evaluatable->False ]
,Cell["
 cm2 = condmap[ {\n { m1, pointerjxa1 == -1/2 }
\n , { m2, pointerjxa1 == +3/2 }
\n } ];
", "Input"]
,Cell["
 cm2 ** qc2", "Input"]
,Cell["\n Similarely, you will get an error if a map acts outside the\n state space of the state it is applied to: ","Text", Evaluatable->False ]
,Cell["
 m2 ** ( qc2 ~contract~ Not[qa1] )", "Input"]
,Cell["\n For a somewhat less arbitrary example of a conditional\n map, have a look at teleporter.vma or teleporter.nb! ","Text", Evaluatable->False ]

}]